(* ::Package:: *)

BeginPackage["MCDist`"]


MC::usage="MC[<transition matrix>, <label list>] is the data structure for defining a Markov chain.";


MCQ::usage="MCQ[M] checks whether M is a well-defined Markov Chain with rational transition matrix";


MCtm::usage="MCtm[<array rules>, <number of states>] returns a transition matrix.";


JoinMC::usage="JoinMC[<Markov chain>,...,<Markov chain>] returns an MC representing the disjoint union of the given sequence of MCs";


PlotMC::usage="PlotMC[M] displays the underlying graph of the Markov chain M.";


BDistMC::usage="BDistMC[<Markov chain>, <discount factor>, <query list>] computes the bisimilarity distance between all the pairs of states in <query list>.";


Verbose::usage="Option for BDistMC (default value False). When the Verbose option is set True the encoutered couplings are displayed during the compuatation";


ConsistencyCheck::usage="Option for BDistMC (default value True). When the ConsistencyCheck is set True, a sanity check of the input is performed before starting the compuatation.";


Estimates::usage="Option for BDistMC (default value {}). It is possible to provide some over-estimates for that will be used as extact values by our method. It can be used to further limit the on-the-fly exploration of the state space during the computation.";


BisimQuotientMC::usage="BisimQuotientMC[<Markov chain>] returns the binimilarity quotient of the given MC, that is, all bisimilar states are lumped together";


BisimClassesMC::usage="BisimClassesMC[<Markov chain>] return the set of states of the given MC partitioned into bisimilarity classes";


RandomMC::usage="RandomMC[<number of states>, <max outer-degree>] randomly generates an MC";


Begin["`Private`"]


(* ::Section:: *)
(*Public*)


(* ::Subsection:: *)
(*Labelled Markov Chains (LMCs)*)


MCQ[MC[TM_,labels_]]:=
Module[{size,a,b},
size=Length@labels; {a,b}=Dimensions[TM];
(* check if the sizes are consistent, and if TM is a probability distribution matrix *)
size==a\[And]a==b\[And] TestAll[(Plus@@#&)/@TM,#==1&]
]/;MatchQ[labels,{__Integer}]\[And] (* labels is a list of integers *)
MatrixQ[TM,MatchQ[#,_Rational]\[Or] IntegerQ[#]&](* TM is a matrix of rationals *);
MCQ[_]:=False;


MCtm[arrayRules_, numStates_]:=SparseArray[arrayRules,{numStates,numStates}];


(* ::Text:: *)
(*JoinMC returns a graphical representation of a given Markov chain *)


JoinMC[seq__MC]:=Fold[JoinMCAux,First@{seq},Rest@{seq}];
JoinMCAux[MC[tm1_,l1_],MC[tm2_,l2_]]:=
Module[{n = Length@l1, m=Length@l2},
MC[ArrayFlatten[{{tm1,SparseArray[{},{n,m}]},{SparseArray[{},{m,n}],tm2}}],
Join[l1,l2]]
];


(* ::Text:: *)
(*PlotMC returns a graphical representation of a given Markov chain *)


PlotMC[MC[tm_,labels_]]:=
Module[{f,g,edgelist},
f[Edge[x_,p_,y_]]:=Labeled[x->y,p];
g[lbl_,{vertex_}]:=vertex->Hue[1/(lbl+1)];
edgelist=Edge[#[[1,1]],#[[2]],#[[1,2]]]&/@Most@ArrayRules[tm];
Graph[f/@edgelist,VertexLabels->"Name",VertexSize->0.1,VertexStyle->MapIndexed[g,labels]]
];


(* ::Subsection:: *)
(*On-the-fly Algorithm for computing Bisimilarity Pseudometric*)


Options[BDistMC]={Verbose->False,ConsistencyCheck->True,Estimates-> {}};
BDistMC::query="The following queries are ill-formed: `1`.";
BDistMC::MarkovChain = "The 1-st argument is not a well-formed rational-valued probabilistic Markov Chain.";
BDistMC::DiscountFactor ="The second argument, namely `1`, must be a rational in the interval (0,1].";
BDistMC::MultipleEstimates="There must be at most one estimate for each pair";
BDistMC::WrongEstimates="Estimates must be a rationals in the interval (0,1]";

BDistMC[mc:MC[_,_],\[Lambda]_, All,options: OptionsPattern[{Verbose -> False,ConsistencyCheck->True,Estimates-> {}}]]:=
BDistMC[mc,\[Lambda],AllPairs@mc,options];
BDistMC[mc:MC[_,_],\[Lambda]_, queryLst_, OptionsPattern[{Verbose -> False,ConsistencyCheck->True,Estimates-> {}}]]:=
Module[{},
If[OptionValue[ConsistencyCheck],CheckInput[mc,\[Lambda], queryLst]];
(* Set Options for ComputeDistance *)
SetOptions[BDistMC,Verbose->OptionValue[Verbose]];
SetOptions[BDistMC,ConsistencyCheck->OptionValue[ConsistencyCheck]];
SetOptions[BDistMC,Estimates->OptionValue[Estimates]];
(* Start computing the Subscript[d, \[Lambda]](s,t) *)
ComputeDistancesAux[mc,\[Lambda],queryLst]
];


(* ::Subsection:: *)
(*Bisimilarity *)


(* ::Text:: *)
(*BisimQuotientMC[mc] returns an MC with all bisimilar states lumped together*)


BisimQuotientMC[mc:MC[tm_,labels_]]:=
Module[{Rate,Quo},
Quo=BisimClassesMC[mc];
Rate[class1_,class2_]:=Plus@@(tm[[First@class1,#]]&/@class2);
MC[SparseArray@Table[Rate[i,j],{i,\!\(TraditionalForm\`Quo\)},{j,Quo}],labels[[First@#]]&/@Quo]
];


(* ::Text:: *)
(*BisimClassesMC[mc] returns the quotient w.r.t. bisimilarity*)


BisimClassesMC[mc:MC[tm_,labels_]]:=
\!\(TraditionalForm\`FixedPoint\)[(Union@@#&)/@Gather[#,Intersection[#1,#2]!= {}&]&,
Select[BDistMC[mc,1,All, Verbose->False],#[[2]]==0&][[All,1]]];


(* ::Section:: *)
(*Private*)


(* ::Subsection::Closed:: *)
(*Utilities*)


DiffDist[d1_,d2_]:=Max[Flatten[Abs[d1-d2]]];


ZeroMatrix[n_]:=SparseArray[{},{n,n}];


TestAll[list_,test_]:=And@@(test/@list);


DebugPrint[s___]:=Print[s]/;OptionValue[BDistMC,Verbose];


PrintActiveProblems[dist_,problems_,queryLst_]:=
Module[{ActivePrint,active = ReachableProblems[dist,problems,queryLst]},
ActivePrint[{u_,v_}->actv_]:=Print[{u,v}->PrPrint[problems[[u,v]]]]/;actv;
Print["Reachable states from ",queryLst, " in the current coupling"];
Scan[ActivePrint,ArrayRules[active]]
(*DebugPrint[MatrixForm@MapIndexed[ActivePrint,problems,{2}]]*)
]/;OptionValue[BDistMC,Verbose];


PrintProblems[problems_]:=
DebugPrint[MatrixForm@Map[PrPrint,problems,{2}]];


(* ::Subsection::Closed:: *)
(*Labelled Markov Chains (LMCs)*)


AllPairs[MC[_,labels_]]:=Module[{size=Length@labels},Join@@(Table[{i,j},{i,size},{j,size}])];


SameLabelsPairs[mc:MC[_,labels_]]:=Select[AllPairs@mc,Equal@@(labels[[#]])&];


(* ::Subsection:: *)
(*Random generation of LMCs*)


RandomRational[rmax_,n_]:=
(RandomInteger[{1,#}]/#)&/@RandomInteger[{1,rmax},n]


(* ::Text:: *)
(*RandomProbDistr gives a pseudorandom list of rationals Subscript[r, 1],..., Subscript[r, n] such that \!\( *)
(*\*UnderoverscriptBox[\(\[Sum]\), \(i = 1\), \(n\)]*)
(*\*SubscriptBox[\(r\), \(i\)]\)= 1*)


RandomProbDistr[rmax_,n_]:=
Module[{rnd=RandomRational[rmax,n],total},
total=Plus@@rnd;
(#/total)&/@rnd
]


RandomTransitionMatrix[numStates_,maxDegree_]:=
Module[{MakeRow},
MakeRow[degree_]:=Shuffle@Join[Table[0,{numStates-degree}],RandomProbDistr[numStates,degree]];
SparseArray@Table[MakeRow[RandomInteger[{1,maxDegree}]],{numStates}]
]/;numStates>= maxDegree;
RandomTransitionMatrix[numStates_,maxDegree_]:=
(Message[RandomTransitionMatrix::nnarg,numStates,maxDegree];Abort[]);


RandomTransitionMatrix::nnarg="The argument `1` is not greater than or equal to `2`.";


RandomMC[numStates_,maxDegree_]:=
MC[RandomTransitionMatrix[numStates,maxDegree],RandomInteger[{0,1},numStates]];


Shuffle[{}]:={};
Shuffle[lst_]:=
Module[{left,right,pick=RandomInteger[{1,Length@lst}]},
left=Take[lst,pick];right=Drop[lst,pick];
Join[{lst[[pick]]},Shuffle[Most@left],Shuffle[right]]
];


(* ::Subsection::Closed:: *)
(*Transportation Problem*)


(* ::Text:: *)
(*Given two \[Epsilon]-numbers x and y, EpsilonLeq[x,y] determines if x <= y provided that \[Epsilon] -> +0.*)


EpsilonLeq[x_,y_]:=
If[EpsilonRemoval[x-y]==0,(* check if x-y depends only on \[Epsilon]*)
EpsilonFix[x-y,1]<= 0,(* set \[Epsilon] > 0 and check again *)
EpsilonRemoval[x-y]<= 0 (* set e[]= 0 and check again *)
];


EpsilonMin[x_,y_]:= If[EpsilonLeq[x,y],x,y];


EpsilonPerturbation[\[Pi]s_,\[Pi]t_]:=
Module[{AddEpsilon},
AddEpsilon[k_][{a_,id_}]:={a + k e[],id};
{AddEpsilon[1]/@\[Pi]s, Append[Most[\[Pi]t],AddEpsilon[Length@\[Pi]s][Last@\[Pi]t]]}
];


EpsilonFix[x_,\[Epsilon]value_]:=x/.e[]->\[Epsilon]value;


EpsilonRemoval[x_]:=EpsilonFix[x,0];


indexed[lst_]:=Thread[{lst,Table[i,{i,1,Length@lst}]}];


GuessVertexAux[_][\[Pi]s_,\[Pi]t_]:={}/;\[Pi]s=={}||\[Pi]t=={};
GuessVertexAux[strategy_][\[Pi]s_,\[Pi]t_]:=
Module[{G},
G[{i_,j_}]:=
Module[{rec},
rec=If[EpsilonLeq[\[Pi]s[[i,1]],\[Pi]t[[j,1]]],(*\[Pi]s\[LeftDoubleBracket]i,1\[RightDoubleBracket] \[LessEqual] \[Pi]t\[LeftDoubleBracket]j,1\[RightDoubleBracket]*)
GuessVertexAux[strategy][Drop[\[Pi]s,{i}],ReplacePart[\[Pi]t,{j,1}->\[Pi]t[[j,1]]- \[Pi]s[[i,1]]]],
GuessVertexAux[strategy][ReplacePart[\[Pi]s,{i,1}->\[Pi]s[[i,1]]- \[Pi]t[[j,1]]],Drop[\[Pi]t,{j}]]];
Prepend[rec,{\[Pi]s[[i,2]],\[Pi]t[[j,2]]}->EpsilonMin[\[Pi]s[[i,1]],\[Pi]t[[j,1]]]]
];
G[strategy[\[Pi]s,\[Pi]t]]
];


NorthWestStrategy[_,_]:={1,1};


ToMatrix[\[Pi]s_,\[Pi]t_,vertex_]:=
Module[{mapS,revmapS,mapT,revmapT,toMap,mapIndex,mappedvertex},
toMap[indexlst_]:=
Module[{index=Table[i,{i,1,Length@indexlst}] },
{Thread[Rule[index,indexlst]],Thread[Rule[indexlst,index]]}];
mapIndex[{s_,t_}->val_]:={s/.revmapS,t/.revmapT}->val;

{{mapS,revmapS},{mapT,revmapT}} = toMap[#[[All,2]]]&/@{\[Pi]s,\[Pi]t};
mappedvertex= mapIndex/@vertex;
IndexedVertex[mapS,mapT,mappedvertex[[All,1]],SparseArray[mappedvertex,Length/@{\[Pi]s,\[Pi]t}]]
];


GuessVertex[\[Pi]s_,\[Pi]t_]:=
Module[{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t},
{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t}=EpsilonPerturbation[\[Pi]s, \[Pi]t];
ToMatrix[\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t,GuessVertexAux[NorthWestStrategy][\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t]]
];


(* ::Text:: *)
(*AddtoBase[d, vertex] == {reducedcost, non-basic cell} *)
(*returns the non-basic cell which has the minimal reduced cost, if the vertex has no non-basic cells, then it returns {\[Infinity], {0,0}}.*)


AddtoBase[dist_,IndexedVertex[mapS_,mapT_,basicCells_,V_]]:=
Module[{sizeS=Length@mapS, sizeT=Length@mapT, 
t,getCost,ReducedCost,uv,makeRow,base},

makeRow[{1,j_}]:={SparseArray[{{1,sizeS+j}->1},{1,sizeS+sizeT}]};
makeRow[{i_,j_}]:={SparseArray[{{1,i}->1,{1,sizeS+j}->1},{1,sizeS+sizeT}]};

getCost[{i_,j_}]:=getDistValue[dist,{i/.mapS,j/.mapT}];

ReducedCost[_,{i_,j_}]:=
Module[{reducedcost=getCost[{i,j}]-uv[[i]]-uv[[sizeS+j]]},
If[base[[1]]>reducedcost,
base={reducedcost,{i,j}} (* update the current least reduced cost *)
];reducedcost];

(* u-v modifiers (Subscript[u, 1],...,Subscript[u, sizeS],Subscript[v, 1],...,Subscript[v, sizeT]) *)
uv=LinearSolve[Normal@ArrayFlatten[makeRow/@basicCells],getCost/@basicCells];
(*DebugPrint["u-v Problem:",MatrixForm@ArrayFlatten[makeRow/@basicCells],
MatrixForm@uv,MatrixForm[getCost/@basicCells]];*)
base={\[Infinity],{0,0}};(* init cell *)
(* scan the cells computing their reduced cost *)
MapIndexed[ReducedCost,V,{2}];
base (* return the cell with the least reduced cost *)
];


(* ::Text:: *)
(*SteppingStone[non-basic cell, vertex] returns the vertex obtained by inserting the given non-basic cell into the base *)


SteppingStone[{s_,t_},vertex:IndexedVertex[mapS_,mapT_,basicCells_,V_]]:=
Module[{M,d,oldBasicCell,newBasicCells},
{oldBasicCell,M}=SteppingStonePath[{s,t},vertex];
(*DebugPrint[MatrixForm@M];*)
(*DebugPrint["old basic cell: ",oldBasicCell, ", new basic cell: ",{s,t}];*)
d= Extract[V,oldBasicCell];
(*DebugPrint["ammount moved: ",d];*)
newBasicCells = basicCells/.oldBasicCell->{s,t};
(*DebugPrint["Old basic cells:",basicCells,", new basic cells",newBasicCells];*)
IndexedVertex[mapS,mapT,newBasicCells,SparseArray[V+d M]]
];


SteppingStonePath[{s_,t_},vertex:IndexedVertex[mapS_,mapT_,basicCells_,V_]]:=
Module[{neighbors,visit,parent,cell,path,getMatrix,visited,f},

visited[_]:=False;
neighbors[{u_,True}]:={#[[2]],False}&/@(Select[basicCells,#[[1]]==u \[And] Not[visited[{#[[2]],False}]]&]);
neighbors[{v_,False}]:={#[[1]],True}&/@(Select[basicCells,#[[2]]==v \[And] Not[visited[{#[[1]],True}]]&]);

f[{{a_,True},{b_,False}}]:={a,b};
f[{{a_,False},{b_,True}}]:={b,a};

visit[node_]:=
(visited[node]=True;
Scan[(parent[#]=node; visit@#)&,neighbors[node]]
);

getMatrix[{},sign_]:={};
getMatrix[lst_,sign_]:=
(If[sign<0\[And] EpsilonLeq[Extract[V,First@lst] ,Extract[V,cell]],cell=First@lst];(* update cell *)
Append[getMatrix[Rest@lst,- sign], First@lst -> sign]); (* compute arrayrules *)

(*DebugPrint[PrPrint@vertex];*)

(* search for an alternating path from s to t *)
parent[{s,True}]={};(* {s,True} is the root of the tree *)
visit[{s,True}]; (* compute the spanning tree rooted in {s,True} *)

(* get the alternating path from {t,False} to {s,True} *)
path=Append[f/@Partition[(Most@NestWhileList[parent,{t,False},#!={}&]),2,1],{s,t}];
cell=First@path; path=getMatrix[path,-1];

{cell,SparseArray[path,Dimensions@V]}
];


BestVertex[dist_,{redCost_,nbCell_},vertex_]:=
Module[{currVertex=vertex, currRedCost=redCost,currNbCell= nbCell},
While[currRedCost<0,
currVertex= SteppingStone[currNbCell,currVertex];(* move to a better vertex *)
(*DebugPrint[PrPrint@currVertex];*)
(* update the current least reduced cost *)
{currRedCost,currNbCell}= AddtoBase[dist,currVertex];
];
currVertex
];


PrPrint[IndexedVertex[mapS_,mapT_,basicCells_,V_]]:=
Module[{basic},
basic[value_,index_]:=Style[EpsilonRemoval@value,Red,Bold]/;MemberQ[basicCells,index];
basic[value_,_]:=value;
TableForm[MapIndexed[basic,V,{2}],TableHeadings->{ToString[#[[2]]]&/@mapS,ToString[#[[2]]]&/@mapT}]
];
PrPrint[x_]:=x;


(* ::Subsection::Closed:: *)
(*Iterative Algorithm*)


(* ::Text:: *)
(*Given the numer of states |S|, BottomDist[|S|] returns the bottom distance*)


BottomDist[size_]:=Table[0,{size},{size}];


TopDist[size_]:=Table[1,{size},{size}];


(* ::Text:: *)
(*The funtion \[CapitalDelta] implements the distance transformer*)


\[CapitalDelta][MC[TM_,labels_],\[Lambda]_][dist_]:=
Module[{size=Length@labels,Marginal,TP,NewValue},
Marginal[state_]:=Select[indexed[TM[[state,All]]],#[[1]]>0&];
TP[s_,t_]:=Module[{vertex=GuessVertex[Marginal[s],Marginal[t]]},
BestVertex[dist,AddtoBase[dist,vertex],vertex]];

NewValue[s_,t_]:=1/;labels[[s]]!= labels[[t]];
NewValue[s_,t_]:=
Module[{cellCost,optSchedule=TP[s,t]},
cellCost[IndexedVertex[mapS_,mapT_,_,V_]][{i_,j_}]:=
getDistValue[dist,{i/.mapS,j/.mapT}]*EpsilonRemoval[V[[i,j]]];
\[Lambda] *Plus@@(cellCost[optSchedule]/@optSchedule[[3]])
];
(* compute the new values for each pair *)
Table[NewValue[s,t],{s,size},{t,size}]
];


IterativeAlg[mc:MC[_,labels_],\[Lambda]_,OptionsPattern[{StoppingTime -> 4,Verbose->False}]]:=
Module[{t=0,iterCount=0,size=Length[labels],dt,old,curr},
SetOptions[BDistMC,Verbose->OptionValue[Verbose]];
(* initialization *)
old=TopDist[Length[labels]];
curr=BottomDist[Length[labels]];
(* iteratively apply \[CapitalDelta], until the extact result is reached or
the StoppingTime has been exceeded  *)
While[t <= OptionValue[StoppingTime] \[And] curr != old,(*  *)
old=curr;
{dt,curr}=Timing[\[CapitalDelta][mc,\[Lambda]][old]];
t+=dt;iterCount++;
];
{t,R[iterCount,Length[SameLabelsPairs[mc]],curr]}
];


(* ::Subsection:: *)
(*On-the-fly Algorithm for computing Bisimilarity Pseudometric*)


(* ::Text:: *)
(*Given as input an indexed vertex v, SubProblems[v] returns the list of active subproblems. This is simply done by selecting the (non-degenerate) basic cells of \[Omega].*)


SubProblems[IndexedVertex[mapS_,mapT_,basicCells_,V_]]:= 
DeleteDuplicates[(Sort[{#[[1]]/.mapS,#[[2]]/.mapT}]&)/@Select[basicCells,EpsilonRemoval[Extract[V,#]]>0&]];


FixedDistance[dist_,{i_,j_}]:=MatchQ[dist[[i,j]],_Fixed] ;(* says if the distance is fixed or not *)


KnownProblem[dist_,problems_,{i_,j_}]:= (* decide whether the problem (i,j) has already been considered *)
FixedDistance[dist,{i,j}]||MatchQ[problems[[i,j]],_IndexedVertex];


AddProblem[mc:MC[_,_],dist_][problems_,{i_,j_}]:= problems/; KnownProblem[dist,problems,{i,j}];
AddProblem[mc:MC[TM_,labels_],dist_][problems_,{i_,j_}]:=
Module[{Marginal,newvertex},
Marginal[state_]:=Select[indexed[TM[[state,All]]],#[[1]]>0&];
newvertex = GuessVertex[Marginal[i],Marginal[j]];(* guess a fesible transposrtation schedule *)
UpdateProblems[mc,dist][problems,{i,j},newvertex](* recursively update the problems matrix *)
];


AddProblems[mc:MC[_,_],dist_][problems_,queryLst_]:=Fold[AddProblem[mc,dist],problems,queryLst];


UpdateProblems[mc:MC[_,_],dist_][problems_,{i_,j_},newvertex_]:=
AddProblems[mc,dist][ReplacePart[problems,{i,j}->newvertex],SubProblems[newvertex]];


VertexCoeffs[MC[_,labels_],dist_,IndexedVertex[mapS_,mapT_,basicCells_,V_]]:=
Module[{coeffValue,gatherFun,aggr},
coeffValue[{i_,j_}]:=getDistValue[dist,{i/.mapS,j/.mapT}]*EpsilonRemoval[V[[i,j]]];

gatherFun[{i_,j_}]:=0/;FixedDistance[dist,Sort[{i/.mapS,j/.mapT}]];(* constant *)
gatherFun[{i_,j_}]:=Sort[{i/.mapS,j/.mapT}];(* coefficient *)

aggr[VCoeffs[unknown_,const_],lst:{x_,___}]:=
VCoeffs[unknown,Fold[#1+coeffValue[#2]&,const,lst]]/;gatherFun[x]==0;
aggr[VCoeffs[unknown_,const_],lst:{x_,___}]:=
VCoeffs[Append[unknown,gatherFun[x]->Fold[#1+EpsilonRemoval[Extract[V,#2]]&,0,lst]],const];

Fold[aggr,VCoeffs[{},0],GatherBy[Select[basicCells,coeffValue[#]>0&],gatherFun]]
]


ActiveProblems[mc:MC[_,labels_],dist_,problems_,queryLst_]:=
Most@ArrayRules@Fold[ActiveProblemsAux[mc,dist,problems],ZeroMatrix[Length@labels],queryLst];
ActiveProblemsAux[MC[_,labels_],dist_,_][active_,{s_,t_}]:=
active/;FixedDistance[dist,{s,t}]||MatchQ[active[[s,t]],_VCoeffs];
ActiveProblemsAux[mc:MC[_,_],dist_,problems_][active_,{s_,t_}]:=
Fold[ActiveProblemsAux[mc,dist,problems],
ReplacePart[active,{s,t}->VertexCoeffs[mc,dist,problems[[s,t]]]],
SubProblems[problems[[s,t]]]];


ReachableProblems[dist_,problems_,queryLst_]:=
Fold[DFS[dist,problems],SparseArray[{},Dimensions@dist,False],queryLst];
DFS[dist_,problems_][visited_,{s_,t_}]:=visited/;visited[[s,t]]\[Or]FixedDistance[dist,{s,t}];
DFS[dist_,problems_][visited_,{s_,t_}]:=
Fold[DFS[dist,problems],ReplacePart[visited,{s,t}->True],SubProblems[problems[[s,t]]]];


UpdateWithDiscrepancy[\[Lambda]_,mc:MC[_,labels_],dist_,problems_,queryLst_]:=
Module[{aggr,V0={},V={},AdjList=Table[{},{Length@labels},{Length@labels}],
eqrules,DFS,notFree,zeroVars,freeVariables,RemoveFree,PrintD,d,
toBind,makeRow,A,b,sol,eqNumber,varMap,invVarMap,constMap},
PrintD[{x_,y_}]:="d("<>ToString[x]<>","<>ToString[y]<>")";

aggr[id_->VCoeffs[lst_,n_]]:=
(If[n==0,V0 =Append[V0,id],V=Append[V,id]];
AdjList=Fold[ReplacePart[#1,#2->Append[Extract[#1,#2],id]]&,AdjList,lst[[All,1]]]);
DFS[visited_,id_]:=visited/;Extract[visited,id];
DFS[visited_,id_]:=Fold[DFS,ReplacePart[visited,id->True],Extract[AdjList,id]];
RemoveFree[id_->VCoeffs[lst_,n_]]:=id->VCoeffs[Select[lst,Extract[notFree,First@#]&],n];

toBind[index_-> VCoeffs[_,const_],{n_}]:={index->n,n->index,n->const};
makeRow[index_-> VCoeffs[rules_,_]]:=
{SparseArray[{1,index/.varMap}-> 1,{1,eqNumber}]- 
\[Lambda] SparseArray[({1,#[[1]]/.varMap}->#[[2]]&)/@rules,{1,eqNumber}]};

eqrules= ActiveProblems[mc,dist,problems,queryLst];(* compute coefficients of active problems *)
(* LOOK FOR POSSIBLE FREE-VARIABLES *)
Scan[aggr,eqrules];(* construct the inverse graph *)

notFree=Fold[DFS,SparseArray[{},Dimensions@dist,False],V];(* detect non-free variables *)
zeroVars=Select[V0,Not@Extract[notFree,#]&]; (* detect zero-variables *)
DebugPrint["Variables detected to be zero:",PrintD/@zeroVars];
(* "GAUSSIAN ELIMINATION" *)
eqrules=RemoveFree/@Select[eqrules,Extract[notFree,First@#]&];(* remove free variables from eqrules *)
sol=
If[(eqNumber=Length@eqrules)>0,
(* CONSTRUCT THE SINGULAR EQUATION SYSTEM *)
{varMap,invVarMap,constMap}=Transpose[MapIndexed[toBind,eqrules]];
A=ArrayFlatten[makeRow/@eqrules];
b=\[Lambda] SparseArray[constMap,{eqNumber}];
sol =LinearSolve[Normal@A,Normal@b];

d=MatrixForm[(PrintD[#/.invVarMap])&/@Table[i,{i,Length@sol}]];
DebugPrint["Reduced Linear System: ",MatrixForm@A,d,"==",MatrixForm@b, 
";  Solution:",d,"=",sol//MatrixForm];

sol,
DebugPrint["Reduced Linear System: it is empty"];
{}
];

ReplacePart[dist,
Join[#->Fixed[0]&/@zeroVars,(* fix free variables to zero *)
MapIndexed[(#2[[1]]/.invVarMap)->#1&,sol]](* new values for the non-free variables *)
]
];


getDistValue[dist_,index:{_,_}]:=
Module[{getValue},
getValue[Fixed[v_]]:=v;
getValue[v_]:=v;

getValue[Extract[dist,Sort@index]]
];


InitDistances[MC[_,labels_],estimates_]:=
Module[{numStates=Length@labels,setDist},
setDist[s_,t_]:=Fixed[0]/;s==t;
setDist[s_,t_]:=Fixed[1]/; labels[[s]]!=labels[[t]];
setDist[s_,t_]:=Fixed[{s,t}/.estimates]/; MemberQ[estimates,{s,t}->_];
setDist[s_,t_]:=1; 

Table[setDist[s,t],{s,numStates},{t,numStates}]
];


(* ::Text:: *)
(*VertexToChange[dist, {s,t}] == {{i,j}, {redCost, nbcell}} if the problem (i,j) can be further minimized, otherwise it returns {}*)


VertexToChange[problems_,dist_,queryLst_]:=
Module[{Q=queryLst,i,j,reducedcost,cell,neighbours,toVisit,
visited=SparseArray[{},Dimensions@problems,False]},

toVisit[{u_,v_}]:=Not[visited[[u,v]]\[Or]FixedDistance[dist,{u,v}]];
(* a new problem should not be visited if it has been visited already or 
its corresponding value is fixed *)
While[Length@Q >0,
(* take a problem from the queue and set it as already visited *)
{i,j}=First@Q;visited=ReplacePart[visited,{i,j}->True];
(*DebugPrint["Visited problem:",{i,j}];*)
(* compute the least reduced cost for the current problem *)
{reducedcost,cell}=AddtoBase[dist,problems[[i,j]]];
If[reducedcost<0,
(* the first encoutered vertex to be changed *)
DebugPrint["Non optimal schedule for: ",{i,j}," reduced cost: ",reducedcost, " for introducing ",cell," as basic"];
Return[{{i,j},{reducedcost,cell}}],
(* otherwise look forward *)
neighbours =Select[SubProblems[problems[[i,j]]],toVisit];
Q = DeleteDuplicates@Join[Rest@Q,neighbours]
]];
(* none of the reached problems can be further decreased *)
{}
];


CheckInput[mc:MC[_,labels_],\[Lambda]_, queryLst_]:=
Module[{BadQuery,badqueries,estimates,BadEstimate},
BadQuery[size_][pair:{_Integer,_Integer}]:=Not[And@@(IntervalMemberQ[Interval[{1,size}],#]&/@pair)];
BadQuery[_][_]:=True;
BadEstimate[size_][pair_->value_]:=BadQuery[size][pair] \[Or] value<0\[Or] value>1\[Or]Not@MatchQ[value,_Rational|1];

estimates = OptionValue[BDistMC,Estimates];
Which[
Not@MCQ[mc],
	Message[BDistMC::MarkovChain];Abort[],
\[Lambda]<=0\[Or] \[Lambda]>1\[Or]Not@MatchQ[\[Lambda],_Rational|1],
	 Message[BDistMC::DiscountFactor,\[Lambda]];Abort[],
Length@estimates >Length@DeleteDuplicates[estimates,First@#1 == First@#2&],
	Message[BDistMC::MultipleEstimates];Abort[],
Select[estimates,BadQuery[Length@labels]]>0,
	Message[BDistMC::WrongEstimates];Abort[],
(badqueries =Select[queryLst,BadQuery[Length@labels]])>0 ,
	Message[BDistMC::pair,badqueries];Abort[]
];
];


ComputeDistancesAux[mc:MC[TM_, labels_], \[Lambda]_, queryLst_]:=
Module[{sortedQuery,query,FixDistance,currProblems, time,
currDist=InitDistances[mc,OptionValue[BDistMC,Estimates]],vtc, prb, rcCell, newvertex,iterations=0},

FixDistance[dist_,{s_,t_}]:=dist/;FixedDistance[dist,{s,t}];
FixDistance[dist_,index:{s_,t_}]:=
Fold[FixDistance,ReplacePart[dist,index->Fixed[getDistValue[dist,index]]],SubProblems[currProblems[[s,t]]]];

(* Initialization *)
time =First@Timing[
sortedQuery=DeleteDuplicates[Sort/@Select[queryLst,Not@FixedDistance[currDist,#]&]];
While[sortedQuery!={},
DebugPrint["Pairs to compute: ",sortedQuery];
(* pick at random a pair from the query *)
query=sortedQuery[[RandomInteger[{1,Length@sortedQuery},1]]];
DebugPrint["Picked pair: ",query];
currProblems = AddProblems[mc,currDist][ZeroMatrix@Length@labels,query];
PrintActiveProblems[currDist,currProblems,query];
currDist =UpdateWithDiscrepancy[\[Lambda], mc,currDist,currProblems,query];
(*DebugPrint[MatrixForm@currDist];*)
(* look for a non-optimal transportation schedule *)
vtc = VertexToChange[currProblems, currDist, query];
While[vtc!= {},
iterations++;
{prb, rcCell} = vtc;(* consider the schedule to be optimized *)
(* move to a better coupling *)
newvertex = BestVertex[currDist, rcCell, Extract[currProblems,prb]];
currProblems = UpdateProblems[mc,currDist][currProblems,prb, newvertex];
PrintActiveProblems[currDist,currProblems,query];
(* update the current distance with the (locally) computed discrepancy *)
currDist = UpdateWithDiscrepancy[\[Lambda], mc,currDist,currProblems,query];
(*DebugPrint[MatrixForm@currDist];*)
(* look for the next non-optimal transportation schedule *)
vtc = VertexToChange[currProblems, currDist, query];
];
(* Set the current active problems as fixed *)
currDist=FixDistance[currDist, First@query];
(*DebugPrint[MatrixForm@currDist];*)
(* update the query list *)
sortedQuery =Select[sortedQuery,Not@FixedDistance[currDist,#]&];
]];
DebugPrint["There are no more pairs to process."];
DebugPrint["Total solved TPs: ",iterations];
DebugPrint["Execution Time: ",time, " sec"];
(#->getDistValue[currDist,#]&)/@queryLst
];


End[]


EndPackage[]
